@@ -31,10 +31,6 @@ struct nftnl_expr_ct {
 #define IP_CT_DIR_ORIGINAL	0
 #define IP_CT_DIR_REPLY		1
 
-#ifndef NFT_CT_MAX
-#define NFT_CT_MAX (NFT_CT_BYTES + 1)
-#endif
-
 static int
 nftnl_expr_ct_set(struct nftnl_expr *e, uint16_t type,
 		       const void *data, uint32_t data_len)
@@ -43,16 +39,16 @@ nftnl_expr_ct_set(struct nftnl_expr *e, uint16_t type,
 
 	switch(type) {
 	case NFTNL_EXPR_CT_KEY:
-		ct->key = *((uint32_t *)data);
+		memcpy(&ct->key, data, sizeof(ct->key));
 		break;
 	case NFTNL_EXPR_CT_DIR:
-		ct->dir = *((uint8_t *)data);
+		memcpy(&ct->dir, data, sizeof(ct->dir));
 		break;
 	case NFTNL_EXPR_CT_DREG:
-		ct->dreg = *((uint32_t *)data);
+		memcpy(&ct->dreg, data, sizeof(ct->dreg));
 		break;
 	case NFTNL_EXPR_CT_SREG:
-		ct->sreg = *((uint32_t *)data);
+		memcpy(&ct->sreg, data, sizeof(ct->sreg));
 		break;
 	default:
 		return -1;
@@ -152,7 +148,7 @@ nftnl_expr_ct_parse(struct nftnl_expr *e, struct nlattr *attr)
 	return 0;
 }
 
-static const char *ctkey2str_array[NFT_CT_MAX] = {
+static const char *ctkey2str_array[NFT_CT_MAX + 1] = {
 	[NFT_CT_STATE]		= "state",
 	[NFT_CT_DIRECTION]	= "direction",
 	[NFT_CT_STATUS]		= "status",
@@ -169,11 +165,19 @@ static const char *ctkey2str_array[NFT_CT_MAX] = {
 	[NFT_CT_LABELS]		= "label",
 	[NFT_CT_PKTS]		= "packets",
 	[NFT_CT_BYTES]		= "bytes",
+	[NFT_CT_AVGPKT]		= "avgpkt",
+	[NFT_CT_ZONE]		= "zone",
+	[NFT_CT_EVENTMASK]	= "event",
+	[NFT_CT_SRC_IP]		= "src_ip",
+	[NFT_CT_DST_IP]		= "dst_ip",
+	[NFT_CT_SRC_IP6]	= "src_ip6",
+	[NFT_CT_DST_IP6]	= "dst_ip6",
+	[NFT_CT_ID]		= "id",
 };
 
 static const char *ctkey2str(uint32_t ctkey)
 {
-	if (ctkey > NFT_CT_MAX)
+	if (ctkey >= NFT_CT_MAX)
 		return "unknown";
 
 	return ctkey2str_array[ctkey];
@@ -218,161 +222,34 @@ static inline int str2ctdir(const char *str, uint8_t *ctdir)
 	return -1;
 }
 
-static int nftnl_expr_ct_json_parse(struct nftnl_expr *e, json_t *root,
-				       struct nftnl_parse_err *err)
-{
-#ifdef JSON_PARSING
-	const char *key_str, *dir_str;
-	uint32_t reg;
-	uint8_t dir;
-	int key;
-
-	if (nftnl_jansson_parse_reg(root, "dreg", NFTNL_TYPE_U32, &reg, err) == 0)
-		nftnl_expr_set_u32(e, NFTNL_EXPR_CT_DREG, reg);
-
-	if (nftnl_jansson_parse_reg(root, "sreg", NFTNL_TYPE_U32, &reg, err) == 0)
-		nftnl_expr_set_u32(e, NFTNL_EXPR_CT_SREG, reg);
-
-	key_str = nftnl_jansson_parse_str(root, "key", err);
-	if (key_str != NULL) {
-		key = str2ctkey(key_str);
-		if (key < 0)
-			return -1;
-
-		nftnl_expr_set_u32(e, NFTNL_EXPR_CT_KEY, key);
-	}
-
-	dir_str = nftnl_jansson_parse_str(root, "dir", err);
-	if (dir_str != NULL) {
-		if (str2ctdir(dir_str, &dir) != 0) {
-			err->node_name = "dir";
-			err->error = NFTNL_PARSE_EBADTYPE;
-			goto err;
-		}
-		nftnl_expr_set_u8(e, NFTNL_EXPR_CT_DIR, dir);
-	}
-
-	return 0;
-err:
-	errno = EINVAL;
-	return -1;
-#else
-	errno = EOPNOTSUPP;
-	return -1;
-#endif
-}
-
-
-static int nftnl_expr_ct_xml_parse(struct nftnl_expr *e, mxml_node_t *tree,
-				      struct nftnl_parse_err *err)
-{
-#ifdef XML_PARSING
-	const char *key_str, *dir_str;
-	int key;
-	uint8_t dir;
-	uint32_t dreg, sreg;
-
-	if (nftnl_mxml_reg_parse(tree, "dreg", &dreg, MXML_DESCEND_FIRST,
-			       NFTNL_XML_OPT, err) == 0)
-		nftnl_expr_set_u32(e, NFTNL_EXPR_CT_DREG, dreg);
-
-	if (nftnl_mxml_reg_parse(tree, "sreg", &sreg, MXML_DESCEND_FIRST,
-			       NFTNL_XML_OPT, err) == 0)
-		nftnl_expr_set_u32(e, NFTNL_EXPR_CT_SREG, sreg);
-
-	key_str = nftnl_mxml_str_parse(tree, "key", MXML_DESCEND_FIRST,
-				     NFTNL_XML_MAND, err);
-	if (key_str != NULL) {
-		key = str2ctkey(key_str);
-		if (key < 0)
-			return -1;
-
-		nftnl_expr_set_u32(e, NFTNL_EXPR_CT_KEY, key);
-	}
-	dir_str = nftnl_mxml_str_parse(tree, "dir", MXML_DESCEND_FIRST,
-				     NFTNL_XML_OPT, err);
-	if (dir_str != NULL) {
-		if (str2ctdir(dir_str, &dir) != 0) {
-			err->node_name = "dir";
-			err->error = NFTNL_PARSE_EBADTYPE;
-			goto err;
-		}
-		nftnl_expr_set_u8(e, NFTNL_EXPR_CT_DIR, dir);
-	}
-
-	return 0;
-err:
-	errno = EINVAL;
-	return -1;
-#else
-	errno = EOPNOTSUPP;
-	return -1;
-#endif
-}
-
 static int
-nftnl_expr_ct_export(char *buf, size_t size, const struct nftnl_expr *e,
-		     int type)
+nftnl_expr_ct_snprintf(char *buf, size_t remain,
+		       uint32_t flags, const struct nftnl_expr *e)
 {
 	struct nftnl_expr_ct *ct = nftnl_expr_data(e);
-	NFTNL_BUF_INIT(b, buf, size);
-
-	if (e->flags & (1 << NFTNL_EXPR_CT_SREG))
-		nftnl_buf_u32(&b, type, ct->sreg, SREG);
-	if (e->flags & (1 << NFTNL_EXPR_CT_DREG))
-		nftnl_buf_u32(&b, type, ct->dreg, DREG);
-	if (e->flags & (1 << NFTNL_EXPR_CT_KEY))
-		nftnl_buf_str(&b, type, ctkey2str(ct->key), KEY);
-	if (e->flags & (1 << NFTNL_EXPR_CT_DIR))
-		nftnl_buf_str(&b, type, ctdir2str(ct->dir), DIR);
-
-	return nftnl_buf_done(&b);
-}
-
-static int
-nftnl_expr_ct_snprintf_default(char *buf, size_t size,
-			       const struct nftnl_expr *e)
-{
-	int ret, len = size, offset = 0;
-	struct nftnl_expr_ct *ct = nftnl_expr_data(e);
+	int ret, offset = 0;
 
 	if (e->flags & (1 << NFTNL_EXPR_CT_SREG)) {
-		ret = snprintf(buf, size, "set %s with reg %u ",
+		ret = snprintf(buf, remain, "set %s with reg %u ",
 				ctkey2str(ct->key), ct->sreg);
-		SNPRINTF_BUFFER_SIZE(ret, size, len, offset);
+		SNPRINTF_BUFFER_SIZE(ret, remain, offset);
 	}
 
 	if (e->flags & (1 << NFTNL_EXPR_CT_DREG)) {
-		ret = snprintf(buf, len, "load %s => reg %u ",
+		ret = snprintf(buf, remain, "load %s => reg %u ",
 			       ctkey2str(ct->key), ct->dreg);
-		SNPRINTF_BUFFER_SIZE(ret, size, len, offset);
+		SNPRINTF_BUFFER_SIZE(ret, remain, offset);
 	}
 
 	if (nftnl_expr_is_set(e, NFTNL_EXPR_CT_DIR)) {
-		ret = snprintf(buf+offset, len, ", dir %s ",
+		ret = snprintf(buf + offset, remain, ", dir %s ",
 			       ctdir2str(ct->dir));
-		SNPRINTF_BUFFER_SIZE(ret, size, len, offset);
+		SNPRINTF_BUFFER_SIZE(ret, remain, offset);
 	}
 
 	return offset;
 }
 
-static int
-nftnl_expr_ct_snprintf(char *buf, size_t len, uint32_t type,
-		       uint32_t flags, const struct nftnl_expr *e)
-{
-	switch (type) {
-	case NFTNL_OUTPUT_DEFAULT:
-		return nftnl_expr_ct_snprintf_default(buf, len, e);
-	case NFTNL_OUTPUT_XML:
-	case NFTNL_OUTPUT_JSON:
-		return nftnl_expr_ct_export(buf, len, e, type);
-	default:
-		break;
-	}
-	return -1;
-}
-
 struct expr_ops expr_ops_ct = {
 	.name		= "ct",
 	.alloc_len	= sizeof(struct nftnl_expr_ct),
@@ -381,7 +258,5 @@ struct expr_ops expr_ops_ct = {
 	.get		= nftnl_expr_ct_get,
 	.parse		= nftnl_expr_ct_parse,
 	.build		= nftnl_expr_ct_build,
-	.snprintf	= nftnl_expr_ct_snprintf,
-	.xml_parse	= nftnl_expr_ct_xml_parse,
-	.json_parse	= nftnl_expr_ct_json_parse,
+	.output		= nftnl_expr_ct_snprintf,
 };
