@@ -22,7 +22,7 @@
 #include <libnftnl/rule.h>
 
 #ifndef NFT_META_MAX
-#define NFT_META_MAX (NFT_META_PRANDOM + 1)
+#define NFT_META_MAX (NFT_META_SDIFNAME + 1)
 #endif
 
 struct nftnl_expr_meta {
@@ -39,13 +39,13 @@ nftnl_expr_meta_set(struct nftnl_expr *e, uint16_t type,
 
 	switch(type) {
 	case NFTNL_EXPR_META_KEY:
-		meta->key = *((uint32_t *)data);
+		memcpy(&meta->key, data, sizeof(meta->key));
 		break;
 	case NFTNL_EXPR_META_DREG:
-		meta->dreg = *((uint32_t *)data);
+		memcpy(&meta->dreg, data, sizeof(meta->dreg));
 		break;
 	case NFTNL_EXPR_META_SREG:
-		meta->sreg = *((uint32_t *)data);
+		memcpy(&meta->sreg, data, sizeof(meta->sreg));
 		break;
 	default:
 		return -1;
@@ -158,6 +158,16 @@ static const char *meta_key2str_array[NFT_META_MAX] = {
 	[NFT_META_OIFGROUP]	= "oifgroup",
 	[NFT_META_CGROUP]	= "cgroup",
 	[NFT_META_PRANDOM]	= "prandom",
+	[NFT_META_SECPATH]	= "secpath",
+	[NFT_META_IIFKIND]	= "iifkind",
+	[NFT_META_OIFKIND]	= "oifkind",
+	[NFT_META_BRI_IIFPVID]	 = "bri_iifpvid",
+	[NFT_META_BRI_IIFVPROTO] = "bri_iifvproto",
+	[NFT_META_TIME_NS]	= "time",
+	[NFT_META_TIME_DAY]	= "day",
+	[NFT_META_TIME_HOUR]	= "hour",
+	[NFT_META_SDIF]		= "sdif",
+	[NFT_META_SDIFNAME]	= "sdifname",
 };
 
 static const char *meta_key2str(uint8_t key)
@@ -181,75 +191,9 @@ static inline int str2meta_key(const char *str)
 	return -1;
 }
 
-static int nftnl_expr_meta_json_parse(struct nftnl_expr *e, json_t *root,
-					 struct nftnl_parse_err *err)
-{
-#ifdef JSON_PARSING
-	const char *key_str;
-	uint32_t reg;
-	int key;
-
-	key_str = nftnl_jansson_parse_str(root, "key", err);
-	if (key_str != NULL) {
-		key = str2meta_key(key_str);
-		if (key >= 0)
-			nftnl_expr_set_u32(e, NFTNL_EXPR_META_KEY, key);
-	}
-
-	if (nftnl_jansson_node_exist(root, "dreg")) {
-		if (nftnl_jansson_parse_reg(root, "dreg", NFTNL_TYPE_U32, &reg,
-					  err) == 0)
-			nftnl_expr_set_u32(e, NFTNL_EXPR_META_DREG, reg);
-	}
-
-	if (nftnl_jansson_node_exist(root, "sreg")) {
-		if (nftnl_jansson_parse_reg(root, "sreg", NFTNL_TYPE_U32, &reg,
-					  err) == 0)
-			nftnl_expr_set_u32(e, NFTNL_EXPR_META_SREG, reg);
-	}
-
-	return 0;
-#else
-	errno = EOPNOTSUPP;
-	return -1;
-#endif
-}
-
-
-static int nftnl_expr_meta_xml_parse(struct nftnl_expr *e, mxml_node_t *tree,
-					struct nftnl_parse_err *err)
-{
-#ifdef XML_PARSING
-	const char *key_str;
-	int key;
-	uint32_t dreg, sreg;
-
-	key_str = nftnl_mxml_str_parse(tree, "key", MXML_DESCEND_FIRST,
-				     NFTNL_XML_MAND, err);
-	if (key_str != NULL) {
-		key = str2meta_key(key_str);
-		if (key >= 0)
-			nftnl_expr_set_u32(e, NFTNL_EXPR_META_KEY, key);
-	}
-
-	if (nftnl_mxml_reg_parse(tree, "dreg", &dreg, MXML_DESCEND_FIRST,
-			       NFTNL_XML_OPT, err) == 0)
-		nftnl_expr_set_u32(e, NFTNL_EXPR_META_DREG, dreg);
-
-	if (nftnl_mxml_reg_parse(tree, "sreg", &sreg, MXML_DESCEND_FIRST,
-			       NFTNL_XML_OPT, err) == 0)
-		nftnl_expr_set_u32(e, NFTNL_EXPR_META_SREG, sreg);
-
-	return 0;
-#else
-	errno = EOPNOTSUPP;
-	return -1;
-#endif
-}
-
 static int
-nftnl_expr_meta_snprintf_default(char *buf, size_t len,
-				 const struct nftnl_expr *e)
+nftnl_expr_meta_snprintf(char *buf, size_t len,
+			 uint32_t flags, const struct nftnl_expr *e)
 {
 	struct nftnl_expr_meta *meta = nftnl_expr_data(e);
 
@@ -264,38 +208,6 @@ nftnl_expr_meta_snprintf_default(char *buf, size_t len,
 	return 0;
 }
 
-static int nftnl_expr_meta_export(char *buf, size_t size,
-				  const struct nftnl_expr *e, int type)
-{
-	struct nftnl_expr_meta *meta = nftnl_expr_data(e);
-	NFTNL_BUF_INIT(b, buf, size);
-
-	if (e->flags & (1 << NFTNL_EXPR_META_DREG))
-		nftnl_buf_u32(&b, type, meta->dreg, DREG);
-	if (e->flags & (1 << NFTNL_EXPR_META_KEY))
-		nftnl_buf_str(&b, type, meta_key2str(meta->key), KEY);
-	if (e->flags & (1 << NFTNL_EXPR_META_SREG))
-		nftnl_buf_u32(&b, type, meta->sreg, SREG);
-
-	return nftnl_buf_done(&b);
-}
-
-static int
-nftnl_expr_meta_snprintf(char *buf, size_t len, uint32_t type,
-			 uint32_t flags, const struct nftnl_expr *e)
-{
-	switch (type) {
-	case NFTNL_OUTPUT_DEFAULT:
-		return nftnl_expr_meta_snprintf_default(buf, len, e);
-	case NFTNL_OUTPUT_XML:
-	case NFTNL_OUTPUT_JSON:
-		return nftnl_expr_meta_export(buf, len, e, type);
-	default:
-		break;
-	}
-	return -1;
-}
-
 struct expr_ops expr_ops_meta = {
 	.name		= "meta",
 	.alloc_len	= sizeof(struct nftnl_expr_meta),
@@ -304,7 +216,5 @@ struct expr_ops expr_ops_meta = {
 	.get		= nftnl_expr_meta_get,
 	.parse		= nftnl_expr_meta_parse,
 	.build		= nftnl_expr_meta_build,
-	.snprintf	= nftnl_expr_meta_snprintf,
-	.xml_parse 	= nftnl_expr_meta_xml_parse,
-	.json_parse 	= nftnl_expr_meta_json_parse,
+	.output		= nftnl_expr_meta_snprintf,
 };
