@@ -70,11 +70,11 @@ int main(int argc, char *argv[])
 		print_err("OOM");
 
 	nftnl_expr_set(ex, NFTNL_EXPR_TG_NAME, "test", strlen("test"));
-	nftnl_expr_set_u32(ex, NFTNL_EXPR_TG_REV, 0x12345678);
+	nftnl_expr_set_u32(ex, NFTNL_EXPR_TG_REV, 0x56781234);
 	nftnl_expr_set(ex, NFTNL_EXPR_TG_INFO, strdup(data), sizeof(data));
 	nftnl_rule_add_expr(a, ex);
 
-	nlh = nftnl_rule_nlmsg_build_hdr(buf, NFT_MSG_NEWRULE, AF_INET, 0, 1234);
+	nlh = nftnl_nlmsg_build_hdr(buf, NFT_MSG_NEWRULE, AF_INET, 0, 1234);
 	nftnl_rule_nlmsg_build_payload(nlh, a);
 
 	if (nftnl_rule_nlmsg_parse(nlh, b) < 0)
